import uuid
from sklearn.utils import check_array, check_X_y, assert_all_finite


class DecisionTree:
    def __init__(self):
        self.results = None
        self.tree_ = None
        self.tree_error_ = None
        self.accuracy_ = None
        self.is_fitted_ = False
        self.statistics = None

    def predict(self):
        pass

    def set_accuracy(self):
        self.accuracy_ = round(
            1 - self.results.error / self.statistics["num_samples"], 5
        )

    @staticmethod
    def is_leaf_node(node):
        return (node["left"] == 0) and (node["right"] == 0)

    def predict(self, X):
        """Implements the standard predict function for a DL8.5 classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample is the label of the closest sample
            seen during fit.
        """

        # Check is fit is called
        # check_is_fitted(self, attributes='tree_') # use of attributes is deprecated. alternative solution is below

        # if hasattr(self, 'sol_size') is False:  # fit method has not been called
        if self.is_fitted_ is False:  # fit method has not been called
            raise NotFittedError(
                "Call fit method first" % {"name": type(self).__name__}
            )

        if self.tree_ is None:
            raise TreeNotFoundError(
                "predict(): ",
                "Tree not found during training by DL8.5 - "
                "Check fitting message for more info.",
            )

        if hasattr(self, "tree_") is False:  # normally this case is not possible.
            raise SearchFailedError(
                "PredictionError: ",
                "DL8.5 training has failed. Please contact the developers "
                "if the problem is in the scope supported by the tool.",
            )

        # Input validation
        X = check_array(X)

        pred = []

        for i in range(X.shape[0]):
            pred.append(self.pred_value_on_dict(X[i, :]))

        return pred

    def pred_value_on_dict(self, instance, tree=None):
        node = tree if tree is not None else self.tree_["tree"][0]
        while not DecisionTree.is_leaf_node(node):
            if instance[node["value"]["test"]] == 1:
                node = self.tree_["tree"][node["right"]]
            else:
                node = self.tree_["tree"][node["left"]]
        return node["value"]["out"]

    def get_dot_body_rec(self, node, parent=None, left=0):
        gstring = ""
        id = str(uuid.uuid4())
        id = id.replace("-", "_")

        if node["right"] == node["left"]:
            gstring += (
                "leaf_"
                + id
                + ' [label="{{class|'
                + str(node["value"]["out"])
                + "}|{error|"
                + str(node["value"]["error"])
                + '}}"];\n'
            )
            gstring += (
                "node_"
                + parent
                + " -> leaf_"
                + id
                + " [label="
                + str(int(left))
                + "];\n"
            )
        else:
            gstring += (
                "node_"
                + id
                + ' [label="{{feat|'
                + str(node["value"]["test"])
                + '}}"];\n'
            )
            gstring += (
                "node_" + parent + " -> node_" + id + " [label=" + str(left) + "];\n"
            )
            gstring += self.get_dot_body_rec(
                self.tree_["tree"][node["left"]], id, left=0
            )
            gstring += self.get_dot_body_rec(
                self.tree_["tree"][node["right"]], id, left=1
            )
        return gstring

    def export_to_graphviz_dot(self):
        gstring = "digraph Tree { \n" "graph [ranksep=0]; \n" "node [shape=record]; \n"
        id = str(uuid.uuid4())
        id = id.replace("-", "_")

        root = self.tree_["tree"][0]
        feat = root["value"]["test"]
        if feat is not None:
            gstring += (
                "node_"
                + id
                + ' [label="{{feat|'
                + str(feat)
                + "}|{error|"
                + str(self.tree_error_)
                + '}}"];\n'
            )
            gstring += self.get_dot_body_rec(
                self.tree_["tree"][root["left"]], parent=id, left=0
            )
            gstring += self.get_dot_body_rec(
                self.tree_["tree"][root["right"]], parent=id, left=1
            )
        gstring += "}"
        return gstring



import json

from .. import *
import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.metrics import DistanceMetric
from sklearn.utils import check_array, check_X_y, assert_all_finite
from pytreesrs.odt import dl85


class DL85Cluster(BaseEstimator, ClusterMixin, DecisionTree):
    def __init__(
        self,
        min_sup=1,
        max_depth=1,
        max_error=1e10,
        max_time=600,
        cache_init_size=0,
        one_time_sort=True,
        lower_bound=ExposedLowerBoundStrategy.None_,
        branching_type=ExposedBranchingStrategy.Dynamic,
        heuristic=ExposedSearchHeuristic.None_,
        cache_init_strategy=ExposedCacheInitStrategy.None_,
        error_function=None,
    ):
        super().__init__()
        self.min_sup = min_sup
        self.max_depth = max_depth
        self.max_error = max_error
        self.max_time = max_time
        self.cache_init_size = cache_init_size
        self.one_time_sort = one_time_sort
        self.data_format = ExposedDataFormat.Tids
        self.specialization = ExposedSpecialization.None_
        self.lower_bound = lower_bound
        self.branching_type = branching_type
        self.heuristic = heuristic
        self.cache_init_strategy = cache_init_strategy
        self.error_function = error_function

        self.results = None

    @staticmethod
    def default_error(tids, X):
        if len(tids) == 0:
            return sys.maxsize, sys.maxsize
        dist = DistanceMetric.get_metric("euclidean")
        X_subset = np.asarray([X[index, :] for index in list(tids)], dtype="int32")
        centroid = np.mean(X_subset, axis=0)
        distances = dist.pairwise(X_subset, [centroid])
        return round(np.sum(distances), 2), DL85Cluster.default_leaf_value(tids, X)

    @staticmethod
    def default_leaf_value(tids, X):
        return round(np.mean(X.take(list(tids))), 2)

    def fit(self, X, X_error=None):

        if X_error is not None:
            assert_all_finite(X_error)
            X_error = check_array(X_error, dtype="int32")

        if self.error_function is None:
            if X_error is None:
                self.error_function = lambda tids: self.default_error(tids, X)
            else:
                if X_error.shape[0] == X.shape[0]:
                    self.error_function = lambda tids: self.default_error(tids, X_error)
                else:
                    raise ValueError(
                        "X_error does not have the same number of rows as X"
                    )

        self.results = dl85(
            X,
            X_error,
            self.min_sup,
            self.max_depth,
            self.max_time,
            self.cache_init_size,
            self.max_error,
            self.one_time_sort,
            self.data_format,
            self.specialization,
            self.lower_bound,
            self.branching_type,
            self.heuristic,
            self.cache_init_strategy,
            self.error_function,
        )

        tree = json.loads(self.results.tree)
        self.statistics = json.loads(self.results.statistics)
        if len(tree["tree"]) == 1 and tree["tree"][0]["value"]["out"] not in [0, 1]:
            self.tree_ = None
        else:
            self.tree_ = tree
            self.is_fitted_ = True
            self.tree_error_ = self.results.error
            self.set_accuracy()



from pytreesrs.enums import (
    ExposedBranchingStrategy,
    ExposedCacheInitStrategy,
    ExposedCacheType,
    ExposedDataFormat,
    ExposedLowerBoundStrategy,
    ExposedSearchHeuristic,
    ExposedSearchStrategy,
    ExposedSpecialization,
)



import json
from .. import *
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils import check_array, check_X_y, assert_all_finite
from pytreesrs.odt import dl85


class DL85Classifier(BaseEstimator, ClassifierMixin, DecisionTree):
    def __init__(
        self,
        min_sup=1,
        max_depth=1,
        max_error=1e10,
        max_time=600,
        cache_init_size=0,
        one_time_sort=True,
        data_format=ExposedDataFormat.ClassSupports,
        specialization=ExposedSpecialization.Murtree,
        lower_bound=ExposedLowerBoundStrategy.Similarity,
        branching_type=ExposedBranchingStrategy.Dynamic,
        heuristic=ExposedSearchHeuristic.None_,
        cache_init_strategy=ExposedCacheInitStrategy.None_,
        error_function=None,
    ):
        super().__init__()
        self.min_sup = min_sup
        self.max_depth = max_depth
        self.max_error = max_error
        self.max_time = max_time
        self.cache_init_size = cache_init_size
        self.one_time_sort = one_time_sort
        self.data_format = data_format
        self.specialization = specialization
        self.lower_bound = lower_bound
        self.branching_type = branching_type
        self.heuristic = heuristic
        self.cache_init_strategy = cache_init_strategy
        self.error_function = error_function

        self.results = None

    def fit(self, X, y=None):

        target_is_need = True if y is not None else False

        if target_is_need:  # target-needed tasks (eg: classification, regression, etc.)
            # Check that X and y have correct shape and raise ValueError if not
            X, y = check_X_y(X, y, dtype="float64")
            # if opt_func is None and opt_pred_func is None:
            #     print("No optimization criterion defined. Misclassification error is used by default.")
        else:  # target-less tasks (clustering, etc.)
            # Check that X has correct shape and raise ValueError if not
            assert_all_finite(X)
            X = check_array(X, dtype="float64")

        self.results = dl85(
            X,
            y,
            self.min_sup,
            self.max_depth,
            self.max_time,
            self.cache_init_size,
            self.max_error,
            self.one_time_sort,
            self.data_format,
            self.specialization,
            self.lower_bound,
            self.branching_type,
            self.heuristic,
            self.cache_init_strategy,
            self.error_function,
        )

        tree = json.loads(self.results.tree)
        self.statistics = json.loads(self.results.statistics)
        if len(tree["tree"]) == 1 and tree["tree"][0]["value"]["out"] not in [0, 1]:
            self.tree_ = None
        else:
            self.tree_ = tree
            self.is_fitted_ = True
            self.tree_error_ = self.results.error
            self.set_accuracy()



import json

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils import check_array, check_X_y, assert_all_finite
from pytreesrs.greedy import lgdt
from .. import ExposedSearchStrategy, DecisionTree


class LGDTCLassifier(BaseEstimator, ClassifierMixin, DecisionTree):
    def __init__(
        self,
        min_sup=1,
        max_depth=2,
        search_strategy=ExposedSearchStrategy.LessGreedyMurtree,
    ):
        super().__init__()
        self.min_sup = min_sup
        self.max_depth = max_depth
        self.search_strategy = search_strategy

    def fit(self, X, y):
        X, y = check_X_y(X, y, dtype="float64")
        self.results = lgdt(
            X,
            y,
            self.search_strategy,
            self.min_sup,
            self.max_depth,
        )

        tree = json.loads(self.results.tree)
        self.statistics = json.loads(self.results.statistics)
        if len(tree["tree"]) == 1 and tree["tree"][0]["value"]["out"] not in [0, 1]:
            self.tree_ = None
        else:
            self.tree_ = tree
            self.is_fitted_ = True
            self.tree_error_ = self.results.error
            self.set_accuracy()



from setuptools import setup, find_packages
from setuptools_rust import Binding, RustExtension

setup(
    version="0.0.1",
    rust_extensions=[RustExtension("pytreesrs", binding=Binding.PyO3)],
    packages=find_packages(),
    # rust extensions are not zip safe, just like C-extensions.
    zip_safe=False,
)
